@@ -147,7 +147,12 @@ class GpgmeWrapper(object):
                     gpgme.gpg_raise_callback_exception(slf)
                 return result
 
-        doc = self._munge_docstring.sub(r'\2.\1(\3', getattr(func, "__doc__"))
+        doc_orig = getattr(func, "__doc__")
+        if doc_orig:
+            doc = self._munge_docstring.sub(r'\2.\1(\3', doc_orig)
+        else:
+            doc = None
+
         _funcwrap.__doc__ = doc
 
         # Monkey-patch the class.
@@ -342,10 +347,12 @@ class Context(GpgmeWrapper):
 
         Decrypt the given ciphertext and verify any signatures.  If
         VERIFY is an iterable of keys, the ciphertext must be signed
-        by all those keys, otherwise an error is raised.  Note: if
-        VERIFY is an empty iterable, that is treated the same as
-        passing verify=True (that is, do verify signatures, but no
-        specific keys are required).
+        by all those keys, otherwise a MissingSignatures error is
+        raised.  Note: if VERIFY is an empty iterable, that is treated
+        the same as passing verify=True (that is, verify signatures
+        and return data about any valid signatures found, but no
+        signatures are required and no MissingSignatures error will be
+        raised).
 
         If the ciphertext is symmetrically encrypted using a
         passphrase, that passphrase can be given as parameter, using a
@@ -353,21 +360,21 @@ class Context(GpgmeWrapper):
         pinentry.
 
         Keyword arguments:
-        sink		-- write result to sink instead of returning it
-        passphrase	-- for symmetric decryption
-        verify		-- check signatures (boolean or iterable of keys,
+        sink            -- write result to sink instead of returning it
+        passphrase      -- for symmetric decryption
+        verify          -- check signatures (boolean or iterable of keys,
                            see above) (default True)
 
         Returns:
-        plaintext	-- the decrypted data (or None if sink is given)
-        result		-- additional information about the decryption
-        verify_result	-- additional information about the signature(s)
+        plaintext       -- the decrypted data (or None if sink is given)
+        result          -- additional information about the decryption
+        verify_result   -- additional information about the valid
+                           signature(s) found
 
         Raises:
         UnsupportedAlgorithm -- if an unsupported algorithm was used
-        BadSignatures	-- if a bad signature is encountered
-        MissingSignatures -- if expected signatures are missing or bad
-        GPGMEError	-- as signaled by the underlying library
+        MissingSignatures    -- if expected signatures are missing or bad
+        GPGMEError           -- as signaled by the underlying library
 
         """
         do_sig_verification = False
@@ -430,13 +437,8 @@ class Context(GpgmeWrapper):
                                               results=results)
 
         if do_sig_verification:
-            # FIXME: should we really throw BadSignature, even if
-            # we've encountered some good signatures?  as above, once
-            # we hit this error, there is no way to accept it and
-            # continue to process the remaining signatures.
-            if any(s.status != errors.NO_ERROR
-                   for s in verify_result.signatures):
-                raise errors.BadSignatures(verify_result, results=results)
+            # filter out all invalid signatures
+            verify_result.signatures = list(filter(lambda s: s.status == errors.NO_ERROR, verify_result.signatures))
             if required_keys is not None:
                 missing = []
                 for key in required_keys:
